import { s } from './h.js'

const COLORS = ['#e22', '#e73', '#fc3', '#ad4', '#4d9', '#3be', '#45d', '#c3e']

/**
 * A helper for creating a chart based on a table (conf. <https://jackens.github.io/nnn/chartable/>).
 *
 * Options:
 * - `bottom`: bottom padding (for X axis labels)
 * - `gapX`: X axis spacing
 * - `gapY`: Y axis spacing
 * - `headerColumn`: flag indicating that `table` has a header column (with X axis labels)
 * - `headerRow`: flag indicating that `table` has a header row (with data series labels)
 * - `id`: chart id
 * - `left`: left padding (for data series labels)
 * - `maxY`: number of Y axis lines
 * - `right`: right padding (for data series labels)
 * - `singleScale`: flag to force single scale
 * - `table`: `HTMLTableElement` to extract data, data series labels and X axis labels
 * - `title`: chart title
 * - `top`: top padding (for the title)
 * - `xLabels`: X axis labels
 * - `zLabels`: data series labels
 * - `zxY`: chart data
 *
 * @param {{
 *   bottom?: number;
 *   gapX?: number;
 *   gapY?: number;
 *   headerColumn?: boolean;
 *   headerRow?: boolean;
 *   id?: string;
 *   left?: number;
 *   maxY?: number;
 *   right?: number;
 *   singleScale?: boolean;
 *   table?: HTMLTableElement;
 *   title?: string;
 *   top?: number;
 *   xLabels?: string[];
 *   zLabels?: string[];
 *   zxY?: number[][];
 * }} options
 */
export const chartable = ({
  bottom = 50,
  gapX = 70,
  gapY = 30,
  headerColumn = false,
  headerRow = true,
  id,
  left = 200,
  maxY = 10,
  right = 200,
  singleScale = false,
  table,
  title,
  top = 70,
  xLabels = [],
  zLabels = [],
  zxY = []
} = {}) => {
  if (table != null) {
    const zxYNotPassed = !zxY.length
    const xLabelsNotPassed = !xLabels.length
    const zLabelsNotPassed = !zLabels.length

    table.querySelectorAll('tr').forEach((row, r) =>
      // @ts-ignore
      row.querySelectorAll('td,th').forEach((/** @type {HTMLTableCellElement} */ col, c) => {
        const x = r - +headerRow
        const z = c - +headerColumn
        const value = col.innerText

        if (x >= 0 && z >= 0 && zxYNotPassed) {
          zxY[z] = zxY[z] ?? []
          zxY[z][x] = parseFloat(value)
        } else if (x >= 0 && z < 0 && xLabelsNotPassed) {
          xLabels[x] = value
        } else if (x < 0 && z >= 0 && zLabelsNotPassed) {
          zLabels[z] = value
        }
      }))
  }

  let bestScales = [Infinity, -Infinity, Infinity, -Infinity, Infinity]

  const ranges = zxY.map(xY => {
    xY = xY.filter(y => !isNaN(y))
    return [Math.min(...xY), Math.max(...xY)]
  })

  if (singleScale) {
    bestScales = ranges.reduce(
      ([totalMin, totalMax], [min, max]) => [Math.min(totalMin, min), Math.max(totalMax, max)],
      bestScales
    )
  } else {
    ranges.forEach(([min1, max0]) => ranges.forEach(([min0, max1]) => {
      if (min0 >= min1 && max1 >= max0) {
        let min2 = Infinity
        let max2 = -Infinity

        ranges.forEach(([min, max]) => {
          if (min < min1 || max1 < max) {
            if (min2 > min) {
              min2 = min
            }
            if (max2 < max) {
              max2 = max
            }
          }
        })

        let cost = 0

        ranges.forEach(([min, max]) => {
          if (min < max) {
            cost += 1 - (max - min) / (min1 <= min && max <= max1 ? max1 - min1 : max2 - min2)
          }
        })

        if (bestScales[4] > cost) {
          bestScales = [min1, max1, min2, max2, cost]
        }
      }
    }))
  }

  const [min1, max1, min2, max2] = bestScales
  const maxX = zxY[0].length
  const xi = Array.from({ length: maxX }, (_, x) => x * gapX)
  const yi = Array.from({ length: maxY }, (_, y) => y * gapY)
  const w = gapX * (maxX - 1)
  const h = gapY * (maxY - 1)
  const svgW = left + w + right
  const svgH = top + h + bottom
  const /** @type {import('./h.js').HArgs1} */ graph = ['g', { transform: `translate(${left} ${top})` },
    ...yi.map(y => ['line', { x1: 0, x2: w, y1: y, y2: y, stroke: '#8888' }]),
    ...xi.map(x => ['line', { x1: x, x2: x, y1: 0, y2: h, stroke: '#8888' }])]
  const /** @type {import('./h.js').HArgs1} */ lColors = ['g', { class: 'chartable-z-colors', transform: 'translate(-80 0)' }]
  const /** @type {import('./h.js').HArgs1} */ lLabels = ['g', { class: 'chartable-z-labels', transform: 'translate(-95 0)' }]
  const /** @type {import('./h.js').HArgs1} */ rColors = ['g', { class: 'chartable-z-colors', transform: 'translate(80 0)' }]
  const /** @type {import('./h.js').HArgs1} */ rLabels = ['g', { class: 'chartable-z-labels', transform: 'translate(95 0)' }]
  const yLabel = (/** @type {number} */ min, /** @type {number} */ max, /** @type {number} */ y) =>
    (min + (max - min) * (maxY - 1 - y) / (maxY - 1)).toFixed(2).replace(/\.?0+$/, '')

  zxY.forEach((xa, z) => {
    const cls = `chartable-z-${z + 1}`
    const fill = COLORS[z % COLORS.length]
    let [min, max] = ranges[z]

    if (min1 <= min && max <= max1) {
      min = min1
      max = max1
      lColors.push(['circle', { class: cls, cx: 0, cy: gapY * (lColors.length - 2), r: 5, fill }])
      lLabels.push(['text',
        { x: 0, y: gapY * (lLabels.length - 2), 'text-anchor': 'end', 'alignment-baseline': 'middle' },
        zLabels[z]])
    } else {
      min = min2
      max = max2
      rColors.push(['circle', { class: cls, cx: 0, cy: gapY * (rColors.length - 2), r: 5, fill }])
      rLabels.push(['text',
        { x: 0, y: gapY * (rLabels.length - 2), 'text-anchor': 'start', 'alignment-baseline': 'middle' },
        zLabels[z]])
    }

    const y = xa.map(x => isNaN(x) ? x : h * (max - x) / (max - min))

    xi.forEach((x, i) => {
      if (!isNaN(y[i])) {
        graph.push(['circle', { class: cls, cx: x, cy: y[i], r: 5, fill }])

        if (!isNaN(y[i - 1])) {
          graph.push(['line', { class: cls, x1: x, x2: xi[i - 1], y1: y[i], y2: y[i - 1], stroke: fill }])
        }
      }
    })
  })

  return s('svg', {
    viewBox: `0 0 ${svgW} ${svgH}`,
    width: `${svgW}px`,
    height: `${svgH}px`,
    class: 'chartable'
  }, id != null ? { id } : null, title != null
    ? ['text',
        { class: 'chartable-title', x: svgW / 2, y: top / 2, 'text-anchor': 'middle', 'alignment-baseline': 'middle' },
        title]
    : null,
  graph,
  min1 < Infinity
    ? ['g', { class: 'chartable-left', transform: `translate(${left} ${top})` },
        ['g', { class: 'chartable-y-labels', transform: 'translate(-15 0)' },
          ...yi.map((y, i) => ['text',
            { x: 0, y, 'text-anchor': 'end', 'alignment-baseline': 'middle' },
            yLabel(min1, max1, i)])],
        lColors, lLabels]
    : null,
  min2 < Infinity
    ? ['g', { class: 'chartable-right', transform: `translate(${left + w} ${top})` },
        ['g', { class: 'chartable-y-labels', transform: 'translate(15 0)' },
          ...yi.map((y, i) => ['text',
            { x: 0, y, 'text-anchor': 'start', 'alignment-baseline': 'middle' },
            yLabel(min2, max2, i)])],
        rColors, rLabels]
    : null,
  xLabels.length > 0
    ? ['g', { transform: `translate(${left} ${top + h})` },
        ['g', { class: 'chartable-x-labels', transform: 'translate(0 15)' },
          ...xi.map((x, i) => ['text',
            { x, y: 0, 'text-anchor': 'middle', 'alignment-baseline': 'hanging' },
            xLabels[i]])]]
    : null
  )
}

export const tests = {}
