/**
 * The type of arguments of the `escapeValues` and `escape` helpers.
 *
 * @typedef {Map<any, (value?: any) => string>} EscapeMap
 */

/**
 * A generic helper for escaping `values` by given `escapeMap`.
 */
export const escapeValues = (
  /** @type {EscapeMap} */ escapeMap,
  /** @type {any[]}     */ values
) => values.map(value => (escapeMap.get(value?.constructor) ?? escapeMap.get(undefined))?.(value) ?? '')

/**
 * A generic helper for escaping `values` by given `escapeMap` (in *TemplateStrings* flavor).
 */
export const escape = (
  /** @type {EscapeMap}            */ escapeMap,
  /** @type {TemplateStringsArray} */ template,
  /** @type {any[]}                */ ...values
) => String.raw(template, ...escapeValues(escapeMap, values))

export const tests = {
  escape: () => {
    // @ts-ignore
    const /** @type {EscapeMap} */ escapeMap = new Map([
      [undefined, () => 'NULL'],
      [Array, (/** @type {any[]} */ values) => escapeValues(escapeMap, values).join(', ')],
      [Boolean, (/** @type {boolean} */ value) => `b'${+value}'`],
      [Date, (/** @type {Date} */ value) => `'${value.toISOString().replace(/^(.+)T(.+)\..*$/, '$1 $2')}'`],
      [Number, (/** @type {number} */ value) => `${value}`],
      [String, (/** @type {string} */ value) => `'${value.replace(/'/g, "''")}'`]
    ])

    // @ts-ignore
    const sql = escape.bind(null, escapeMap)

    const actual = sql`
      SELECT *
      FROM table_name
      WHERE column_name IN (${[true, null, undefined, 42, '42', "4'2", /42/, new Date(323325000000)]})`

    const expected = `
      SELECT *
      FROM table_name
      WHERE column_name IN (b'1', NULL, NULL, 42, '42', '4''2', NULL, '1980-03-31 04:30:00')`

    console.assert(actual === expected)
  }
}
