import { h } from './h.js'
import { has } from './has.js'

const TAGS_TO_SKIP = { IFRAME: 1, NOSCRIPT: 1, PRE: 1, SCRIPT: 1, STYLE: 1, TEXTAREA: 1 }

/**
 * A helper that implements typographic corrections specific to Polish typography.
 */
export const fixTypography = (/** @type {Node} */ node) => {
  const /** @type {Node[]} */ queue = [node]

  while (queue.length > 0) {
    const node0 = queue.shift()

    if (node0 instanceof Element) {
      node0.childNodes.forEach(childNode => {
        if (childNode instanceof Text) {
          queue.push(childNode)
        } else if (childNode instanceof Element && !has(childNode.tagName, TAGS_TO_SKIP)) {
          queue.push(childNode)
        }
      })
    } else if (node0 instanceof Text) {
      const nodeValue = node0.nodeValue?.trim?.()

      if (nodeValue != null) {
        let /** @type {Node} */ previousNode = node0

        nodeValue.split(/(\s|\(|„)([aiouwz—]\s)/gi).forEach((chunk, i) => {
          i %= 3

          const currentNode = i === 2
            ? h('span', { style: 'white-space:nowrap' }, chunk)
            : i === 1
              ? document.createTextNode(chunk)
              : document.createTextNode(chunk.replace(/(\/(?=[^/\s])|\.(?=[^\s]))/g, '$1\u200B'))

          if (node0.parentNode != null) {
            node0.parentNode.insertBefore(currentNode, previousNode.nextSibling)
          }

          previousNode = currentNode
        })

        node0.parentNode?.removeChild(node0)
      }
    }
  }
}

export const tests = {
  fixTypography: () => {
    const p = h('p', 'Pchnąć w tę łódź jeża lub ośm skrzyń fig (zob. https://pl.wikipedia.org/wiki/Pangram).')

    fixTypography(p)

    console.assert(p.innerHTML ===
      'Pchnąć <span style="white-space:nowrap">w </span>tę łódź jeża lub ośm skrzyń fig ' +
      '(zob. https://\u200Bpl.\u200Bwikipedia.\u200Borg/\u200Bwiki/\u200BPangram).')
  }
}
