/**
 * The type of arguments of the `h` and `s` helpers.
 */
export type HArgs = [string | Node, ...HArgs1[]];

/**
 * The type of arguments of the `h` and `s` helpers.
 */
export type HArgs1 = Record<string, any> | null | undefined | Node | string | number | [string | Node, ...HArgs1[]];

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `HTMLElement`s (see also `s`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttribute`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttribute`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `HArgs` are passed to `h` and the results are appended to the element being created or modified.
 */
export function h<T extends keyof HTMLElementTagNameMap>(tag: T, ...args1: HArgs1[]): HTMLElementTagNameMap[T];
export function h<N extends Node>(node: N, ...args1: HArgs1[]): N;
export function h(tagOrNode: string | Node, ...args1: HArgs1[]): Node;

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `SVGElement`s (see also `h`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttributeNS`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttributeNS`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `HArgs` are passed to `s` and the results are appended to the element being created or modified.
 */
export function s<T extends keyof SVGElementTagNameMap>(tag: T, ...args1: HArgs1[]): SVGElementTagNameMap[T];
export function s<N extends Node>(node: N, ...args1: HArgs1[]): N;
export function s(tagOrNode: string | Node, ...args1: HArgs1[]): Node;

export const tests: {
    h: () => void;
    'h: innerText vs items': () => void;
    'h: style': () => void;
    'h: attributes vs properties': () => void;
    'h: nested properties': () => void;
};
