/**
 * A helper that checks if the given argument is of a certain type.
 *
 * @template {abstract new (...args: any[]) => any} T
 *
 * @type {{
 *   (type: BigIntConstructor,  arg: any): arg is bigint;
 *   (type: BooleanConstructor, arg: any): arg is boolean;
 *   (type: NumberConstructor,  arg: any): arg is number;
 *   (type: StringConstructor,  arg: any): arg is string;
 *   (type: SymbolConstructor,  arg: any): arg is symbol;
 *   (type: undefined,          arg: any): arg is undefined | null;
 *   <T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;
 * }}
 *
 * @return {arg is bigint | boolean | number | string | symbol | undefined | null | InstanceType<T>}
 */
export const is = (/** @type {T} */ type, /** @type {any} */ arg) => arg?.constructor === type

export const tests = {
  is: () => {
    console.assert(is(Number, 42))
    console.assert(is(Number, Number(42)))
    // eslint-disable-next-line no-new-wrappers
    console.assert(is(Number, new Number(42)))
    console.assert(is(Number, NaN))
    console.assert(is(String, '42'))
    console.assert(is(String, String('42')))
    // eslint-disable-next-line no-new-wrappers
    console.assert(is(String, new String('42')))
    console.assert(is(Symbol, Symbol('42')))
    console.assert(is(Symbol, Object(Symbol('42'))))
    console.assert(is(undefined, undefined))
    console.assert(is(undefined, null))
    console.assert(is(Object, {}))
    console.assert(is(Array, []))
    console.assert(is(RegExp, /42/))
    console.assert(is(Date, new Date(42)))
    console.assert(is(Set, new Set(['42', 42])))
    console.assert(is(Map, new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]])))
  },

  'is: toString.call': () => {
    const iz = (/** @type {any} */ type, /** @type {any} */ arg) => ({}).toString.call(arg).slice(8, -1) === type?.name

    class FooBar { }

    console.assert(is(FooBar, new FooBar()))
    console.assert(!iz(FooBar, new FooBar()))

    console.assert(!is(Object, new FooBar()))
    console.assert(iz(Object, new FooBar()))

    const fakeFooBar = {}

    // @ts-ignore
    fakeFooBar[Symbol.toStringTag] = FooBar.name

    console.assert(!is(FooBar, fakeFooBar))
    console.assert(iz(FooBar, fakeFooBar))

    console.assert(is(Object, fakeFooBar))
    console.assert(!iz(Object, fakeFooBar))
  }
}
