import { is } from './is.js'

/**
 * The type of arguments of the `jcss` helper.
 *
 * @typedef {{
 *   [attributeOrSelector: string]: string | number | JcssNode | undefined;
 * }} JcssNode
 */

/**
 * The type of arguments of the `jcss` helper.
 *
 * @typedef {Partial<Record<string, JcssNode>>} JcssRoot
 */

const _jcss = (
  /** @type {JcssNode} */ node,
  /** @type {string}   */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  const /** @type {[JcssNode | string[], string][]} */ queue = [[node, prefix]]

  while (queue.length > 0) {
    const [style0, prefix0] = queue.shift() ?? []

    if (style0 == null || prefix0 == null) {
      continue
    }

    if (is(Array, style0)) {
      result.push(prefix0, prefix0 !== '' ? '{' : '', style0.join(';'), prefix0 !== '' ? '}' : '')
    } else {
      const /** @type {[JcssNode | string[], string][]} */ todo = []
      let /** @type {string[]} */ attributes = []
      let attributesPushed = false

      for (const key in style0) {
        const value = style0[key]

        if (is(String, value) || is(Number, value)) {
          if (!attributesPushed) {
            attributesPushed = true
            attributes = []
            todo.push([attributes, prefix0])
          }

          attributes.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else if (value != null) {
          attributesPushed = false

          const /** @type {string[]} */ prefixN = []
          const keyChunks = key.split(',')

          prefix0.split(',').forEach(prefixChunk => keyChunks.forEach(keyChunk => prefixN.push(prefixChunk + keyChunk)))

          todo.push([value, prefixN.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

/**
 * A simple CSS-in-JS helper.
 *
 * The `root` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export const jcss = (/** @type {JcssRoot} */ root, splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ chunks = []

  for (const key in root) {
    const value = root[key]

    if (value != null) {
      if (key[0] === '@') {
        chunks.push(split(key) + '{')
        _jcss(value, '', chunks, split)
        chunks.push('}')
      } else {
        _jcss(value, split(key), chunks, split)
      }
    }
  }

  return chunks.join('')
}

export const tests = {
  'jcss: #1': () => {
    const actual = jcss({
      a: {
        color: 'red',
        margin: 1,
        '.c': { margin: 2, padding: 2 },
        padding: 1
      }
    })

    const expected = `
    a{
      color:red;
      margin:1
    }
    a.c{
      margin:2;
      padding:2
    }
    a{
      padding:1
    }`.replace(/\n\s*/g, '')

    console.assert(actual === expected)
  },

  'jcss: #2': () => {
    const actual = jcss({
      a: {
        '.b': {
          color: 'red',
          margin: 1,
          '.c': { margin: 2, padding: 2 },
          padding: 1
        }
      }
    })

    const expected = `
    a.b{
      color:red;
      margin:1
    }
    a.b.c{
      margin:2;
      padding:2
    }
    a.b{
      padding:1
    }`.replace(/\n\s*/g, '')

    console.assert(actual === expected)
  },

  'jcss: #3': () => {
    const actual = jcss({
      '@font-face$$1': {
        fontFamily: 'Jackens',
        src$$1: 'url(otf/jackens.otf)',
        src$$2: "url(otf/jackens.otf) format('opentype')," +
          "url(svg/jackens.svg) format('svg')",
        fontWeight: 'normal',
        fontStyle: 'normal'
      },
      '@font-face$$2': {
        fontFamily: 'C64',
        src: 'url(fonts/C64_Pro_Mono-STYLE.woff)'
      },
      '@keyframes spin': {
        '0%': { transform: 'rotate(0deg)' },
        '100%': { transform: 'rotate(360deg)' }
      },
      div: {
        border: 'solid red 1px',
        '.c1': { 'background-color': '#000' },
        ' .c1': { backgroundColor: 'black' },
        '.c2': { backgroundColor: 'rgb(0,0,0)' }
      },
      '@media(min-width:200px)': {
        div: { margin: 0, padding: 0 },
        span: { color: '#000' }
      }
    })

    const expected = `
    @font-face{
      font-family:Jackens;
      src:url(otf/jackens.otf);
      src:url(otf/jackens.otf) format('opentype'),url(svg/jackens.svg) format('svg');
      font-weight:normal;
      font-style:normal
    }
    @font-face{
      font-family:C64;
      src:url(fonts/C64_Pro_Mono-STYLE.woff)
    }
    @keyframes spin{
      0%{
        transform:rotate(0deg)
      }
      100%{
        transform:rotate(360deg)
      }
    }
    div{
      border:solid red 1px
    }
    div.c1{
      background-color:#000
    }
    div .c1{
      background-color:black
    }
    div.c2{
      background-color:rgb(0,0,0)
    }
    @media(min-width:200px){
      div{
        margin:0;
        padding:0
      }
      span{
        color:#000
      }
    }`.replace(/\n\s*/g, '')

    console.assert(actual === expected)
  },

  'jcss: #4': () => {
    const actual = jcss({
      a: {
        '.b,.c': {
          margin: 1,
          '.d': {
            margin: 2
          }
        }
      }
    })

    const expected = `
    a.b,a.c{
      margin:1
    }
    a.b.d,a.c.d{
      margin:2
    }`.replace(/\n\s*/g, '')

    console.assert(actual === expected)
  },

  'jcss: #5': () => {
    const actual = jcss({
      '.b,.c': {
        margin: 1,
        '.d': {
          margin: 2
        }
      }
    })

    const expected = `
    .b,.c{
      margin:1
    }
    .b.d,.c.d{
      margin:2
    }`.replace(/\n\s*/g, '')

    console.assert(actual === expected)
  },

  'jcss: #6': () => {
    const actual = jcss({
      '.a,.b': {
        margin: 1,
        '.c,.d': {
          margin: 2
        }
      }
    })

    const expected = `
    .a,.b{
      margin:1
    }
    .a.c,.a.d,.b.c,.b.d{
      margin:2
    }`.replace(/\n\s*/g, '')

    console.assert(actual === expected)
  }
}
