import { eq } from './eq.js'
import { has } from './has.js'
import { is } from './is.js'

/**
 * `JSON.parse` with “JavaScript turned on”.
 *
 * Objects having *exactly* one property which is present in the `handlers` map, i.e. objects of the form:
 *
 * ```js
 * { "«handlerName»": [«params»] }
 * ```
 *
 * are replaced by the result of call
 *
 * ```js
 * handlers['«handlerName»'](...«params»)
 * ```
 */
export const jsOnParse = (
  /** @type {Partial<Record<string, Function>>} */ handlers,
  /** @type {string}                   */ text
) => JSON.parse(text, (key, value) => {
  if (is(Object, value)) {
    let isSecondKey = false

    for (key in value) {
      if (isSecondKey) {
        return value
      }
      isSecondKey = true
    }

    if (has(key, handlers) && is(Array, value[key])) {
      // @ts-ignore
      return handlers[key](...value[key])
    }
  }

  return value
})

export const tests = {
  jsOnParse: () => {
    const handlers = {
      $hello: (/** @type {string} */ name) => `Hello ${name}!`,
      $foo: () => 'bar'
    }
    const actual = jsOnParse(handlers, `[
      {
        "$hello": ["World"]
      },
      {
        "nested": {
          "$hello": ["nested World"]
        },
        "one": 1,
        "two": 2
      },
      {
        "$foo": []
      },
      {
        "$foo": ["The parent object does not have exactly one property!"],
        "one": 1,
        "two": 2
      }
    ]`)
    const expected = [
      'Hello World!',
      {
        nested: 'Hello nested World!',
        one: 1,
        two: 2
      },
      'bar',
      {
        $foo: ['The parent object does not have exactly one property!'],
        one: 1,
        two: 2
      }
    ]

    console.assert(eq(actual, expected))
  }
}
