import { is } from './is.js'

/**
 * A helper that provides information about the given `refs`.
 *
 * It returns an array of triples: `[«name», «prototype-name», «array-of-own-property-names»]`.
 */
export const refsInfo = (/** @type {any[]} */ ...refs) => {
  const /** @type {Set<Function>} */ fns = new Set()

  refs.forEach(ref => {
    while (is(Function, ref) && !fns.has(ref) && `${ref}`.includes('[native code]')) {
      fns.add(ref)
      ref = Object.getPrototypeOf(ref)
    }
  })

  return Array.from(fns.values()).map(/** @return {[string, string, string[]]} */ fn => [
    fn.name,
    Object.getPrototypeOf(fn)?.name ?? '',
    Object.getOwnPropertyNames(fn.prototype ?? Object.create(null)).sort()
  ]).sort((a, b) => -(a[0] < b[0]))
}

export const tests = {
  refsInfo: () => {
    const info = refsInfo(Array, Function)

    console.assert(info.find(([name]) => name === 'Array')?.[2]?.includes?.('length'))
    console.assert(info.find(([name]) => name === 'Function')?.[2]?.includes?.('length'))
  },

  'refsInfo: browserFingerprint': () => {
    const browserFingerprint = () => {
      // @ts-ignore
      const refs = Object.getOwnPropertyNames(window).map(name => window[name])
      const info = refsInfo(...refs)
      const json = JSON.stringify(info)
      const hash = Array(32).fill(0)
      let j = 0

      for (let i = 0; i < json.length; i++) {
        let charCode = json.charCodeAt(i)

        while (charCode > 0) {
          hash[j] = hash[j] ^ (charCode & 15)
          charCode >>= 4
          j = (j + 1) & 31
        }
      }

      return hash.map(x => x.toString(16)).join('')
    }

    console.log(browserFingerprint())
  }
}
